import requests


class EventMessenger:

    """
    Класс для отправки событий в чат ТГ
    """

    def __init__(self, url: str, secret: str = None, host: str = None, chat_id: str = None):
        """
        :param url Ссылка на облачную функцию
        :param secret Секретный ключ
        :param host Хост
        :param chat_id Идентификатор чата
        """
        self.url = url
        self.secret = secret
        self.host = host
        self.chat_id = chat_id
        super().__init__()

    def send_message(self, message: str):
        """
        Отправка сообщения

        :param message Тело сообщения
        """
        json = {
            'secret': self.secret,
            'chat_id': self.chat_id,
            'message': message
        }
        if self.host:
            json['host'] = self.host
        requests.post(self.url, json=json)



import unittest


class CommonTestCase(unittest.TestCase):
    client = None
    app_context = None
    maxDiff = None
    test_docs = []
    authorized = False

    @classmethod
    def setUpClass(cls):
        """Запуск Flask приложения и создание тестовых данных"""
        cls.app = create_app(TestConfig)
        cls.client = cls.app.test_client()
        cls.app_context = cls.app.app_context()
        cls.app_context.push()

        # Создание тестовых данных
        for doc in cls.test_docs:
            doc.save()

    @classmethod
    def tearDownClass(cls):
        """Delete test data and stop Flask app"""
        for doc in cls.test_docs:
            doc.delete()

        # Завершение Flask приложения
        cls.app_context.pop()

    def auth(self, model, email=None, password=None, create_test_user=False, auth_url='/api/login/'):
        """Auth func"""
        auth_url = auth_url

        email = "unit@test.ru" if not email else email
        first_name = "test_user" if not email else email
        password = "test_pass" if not password else password

        auth_user = model.get_by_email(email)
        test_auth_user = None
        if create_test_user and not auth_user:
            test_auth_user = model.objects.create(email=email, first_name=first_name, state='active')
            test_auth_user.set_password(password)
            test_auth_user.save()
        json = {"email": email, "password": password}
        response = self.client.post(auth_url, json=json)
        self.authorized = True if response.status_code == 200 else False
        if test_auth_user:
            self.test_docs.append(test_auth_user)

    def invalid_doc_id(self, url, method, error_type, id_in='url', _id=None):
        args = (url,)
        params = {}
        if id_in == 'data':
            params['json'] = {"id": _id}
        response = method(*args, **params)
        json = self.check_response(response, 400)
        self.assertIn('errors', json)
        self.assertIn('id', json['errors'])
        if error_type == 'not_found':
            self.assertEqual(json['errors']['id'], 'Document not found')
        elif error_type == 'invalid_id':
            self.assertEqual(json['errors']['id'], 'Invalid id')

    def create_success(self, url, model, field_name, data, check_field=True):
        response = self.client.post(url, json=data)
        json = self.check_response(response, 201)
        model = model.objects.filter(id=json['id']).first()
        self.assertNotEqual(model, None)
        self.test_docs.append(model)
        if check_field:
            self.assertEqual(getattr(model, field_name), data[field_name])
        return model

    def create_failed(self, url, field_name, data, valid_type):
        for invalid_param in self.generate_bad_data(valid_type=valid_type):
            data[field_name] = invalid_param
            response = self.client.post(url, json=data)
            json = self.check_response(response, 400)
            self.assertIn('errors', json)
            self.assertIn(field_name, json['errors'])

    def put_success(self, url, edit_obj, edit_field, new_value, check_new_value=True):
        """Success PUT request"""
        data = {edit_field: new_value}
        response = self.client.put(url, json=data)
        json = self.check_response(response)
        self.assertIn('status', json)
        self.assertEqual('success', json['status'])
        edit_obj.reload()
        if check_new_value:
            self.assertEqual(getattr(edit_obj, edit_field), new_value)

    def put_failed(self, url, edit_field=None, bad_data=None):
        """Check error PUT request"""
        if bad_data is None:
            bad_data = []
        for invalid_param in bad_data:
            data = {edit_field: invalid_param}
            response = self.client.put(url, json=data)
            json = self.check_response(response, 400)
            self.assertIn('errors', json)
            self.assertIn(edit_field, json['errors'])

    def delete_success(self, url, delete_obj):
        """Success DELETE request"""
        response = self.client.delete(url, json={"id": delete_obj.id})
        json = self.check_response(response)
        self.assertIn('status', json)
        self.assertEqual('success', json['status'])
        self.assertEqual(getattr(delete_obj, "state"), "hidden")

    def delete_failed(self, url, bad_data, not_found_doc=False):
        """Check error DELETE request"""
        for invalid_param in bad_data:
            response = self.client.delete(url, json={"id": invalid_param})
            json = self.check_response(response, 400)
            self.assertIn('errors', json)
            self.assertIn("id", json['errors'])
            if not_found_doc:
                self.assertEqual(['Неверный формат идентификатора'], json['errors']['id'])

    def check_response(self, response, status_code=200):
        self.assertEqual(response.status_code, status_code)
        self.assertTrue(response.is_json)
        try:
            return response.json
        except Exception:
            self.assertTrue(False)
            return None

    def validate(self, response_json, schema):
        """Validate json response"""
        self.assertIsNotNone(response_json)
        validation_errors = schema(unknown='exclude').validate(response_json)
        if validation_errors:
            print(f"Ошибки при валидации ответа: \n{validation_errors}")
        self.assertDictEqual(validation_errors, {})

    def generate_bad_data(self, valid_type=None, max_length=None, min_length=None):
        self.assertIsNotNone(valid_type)
        invalid_data_map = {
            int: [None, True, "", {}, [], "string", "string1", {"key": "value"}, ["item1"], [1, 2], 1.45],
            float: [None, True, "", {}, [], "string", "string1", {"key": "value"}, ["item1"], [1, 2]],
            str: [None, True, {}, [], 1, {"key": "value"}, ["item1"], [1, 2]],
            bool: [None, "", {}, [], 123, "string", "string1", {"key": "value"}, ["item1"], [1, 2], 1.45],
            list: [None, "", {}, 123, "string", "string1", {"key": "value"}, 1.45],
            "date": [None, True, {}, [], 1, "string", {"key": "value"}, ["item1"], [1, 2], '2020-01-01 10:10'],
        }
        bad_data = invalid_data_map[valid_type]

        # TODO Сделать более универсальным max_length min_length
        if max_length is not None:
            bad_item = ""
            for item in range(max_length + 1):
                bad_item += "s"
            bad_data.append(bad_item)

        if min_length is not None:
            bad_data.append(0)

        return bad_data



import logging
from typing import List, Optional

import boto3
from botocore.exceptions import ClientError, ParamValidationError

__all__ = ["ObjectStorage", "BucketClient"]


date_format = '%Y-%m-%d %H:%M:%S'
log_format = '[%(asctime)s %(levelname)s] %(message)s'
logging.basicConfig(level=logging.INFO, format=log_format, datefmt=date_format)
logger = logging.getLogger(__name__)


class ObjectStorage:
    """ Module for managing the ObjectStorage service """

    def __init__(self, key_id: str, secret_key: str):
        """
        Class initialization

        :param key_id ID of the access key
        :param secret_key Secret key
        """
        self.authorized = False
        self.KEY_ID = key_id
        self.SECRET_KEY = secret_key

        self.s3 = boto3.session.Session().client(service_name='s3', aws_access_key_id=key_id,
                                                 endpoint_url='https://storage.yandexcloud.net',
                                                 aws_secret_access_key=secret_key)
        logger.info("Start authorization ...")
        self.check_auth()

    @staticmethod
    def get_bucket_client(key_id: str, secret_key: str, bucket_name: str) -> "BucketClient":
        """
        Get bucket client

        :param key_id Key identifier
        :param secret_key Secret key
        :param bucket_name Name of the bucket to work with

        :return BucketClient instance
        """
        return BucketClient(key_id, secret_key, bucket_name)

    def check_auth(self) -> bool:
        """
        Check authorization state

        To check authorization, use the `s3.list_buckets()` method
        """
        try:
            self.s3.list_buckets()
        except ClientError:
            self.authorized = False
            logger.error("Authorization failed!")
        self.authorized = True
        logger.info("Authorization success!")
        return self.authorized


class BucketClient(ObjectStorage):

    bucket_name = None

    def __init__(self, key_id: str, secret_key: str, bucket_name: str):
        """
        Class initialization

        :param key_id Key identifier
        :param secret_key Secret key
        :param bucket_name Name of the bucket to work with
        """
        super().__init__(key_id, secret_key)
        self.bucket_name = bucket_name

    def get_list_objects_generator(self, path: str = '', start_after: str = '') -> List[dict]:
        """
        Get list objects from bucket. (Generator)

        :param path Path
        :param start_after Return all elements after the specified value

        :yield List of dictionaries containing information about objects

                Example response: [
                    {
                        'Key': 'banner_1.jpg',
                        'LastModified': datetime.datetime(2020, 6, 12, 12, 37, 2, 686000, tzinfo=tzutc()),
                        'ETag': '"a440b0e3bcdd651c18626df3374463cd"',
                        'Size': 359546,
                        'StorageClass': 'STANDARD'
                    },
                    ...,
                    ...
                ]

        """
        retry = True
        while retry:
            try:
                res = self.s3.list_objects_v2(Bucket=self.bucket_name, StartAfter=start_after, Prefix=path)
            except ClientError as exc:
                self.__validate_client_error(exc)
                break
            retry = res['KeyCount'] == 1000
            start_after = res.get('Contents', [])[-1].get('Key') if retry else ''
            yield res.get('Contents', [])

    def get_list_all_objects(self, path: str = '', limit: int = None) -> List[dict]:
        """
        Get list all objects from bucket

        :param path Path
        :param limit Response limit
        """
        list_objects = []
        logger.debug(f'Getting list all objects from bucket: {self.bucket_name}. Path: {path}')
        for result in self.get_list_objects_generator(path):
            list_objects += result
            if limit and len(result) > limit:
                break
        logger.debug(f'Received a list with information about {len(list_objects[:limit])} files.')
        return list_objects[:limit]

    def get_bytes_object(self, object_name: str) -> Optional[bytes]:
        """
        Get bytes object.

        :param object_name Object name in bucket
        :return Object format bytes or None
        """
        logger.debug(f'Try getting object: {object_name}')
        try:
            object_content = self.s3.get_object(Bucket=self.bucket_name, Key=object_name)
        except ClientError as exc:
            return self.__validate_client_error(exc)
        if object_content['ContentType'] == 'application/x-directory':
            return None
        logger.debug('File getting successfully.')
        return object_content['Body'].read()

    def upload_file(self, file_name: str, file_content: bytes):
        """
        Upload file to bucket

        :param file_name File name
        :param file_content File content
        """
        logger.debug(f'Try upload file to path: {file_name}')
        try:
            return self.s3.put_object(Bucket=self.bucket_name, Key=file_name, Body=file_content)
        except (ClientError, ParamValidationError) as exc:
            if exc.__class__.__name__ == 'ParamValidationError':
                return logger.error(f'{exc.args[0]}')
            return self.__validate_client_error(exc)

    def move_objects(self, out_path: str, objects_list: Optional[List[dict]], in_path: Optional[str]):
        """
        Move objects.

        :param out_path: Path to move files to
        :param objects_list: List of objects to move. (Received by the `self.get_all_list()` method)
        :param in_path: Path to move files from
        :return None
        """
        def move_objects_list(objects):
            """
            Move objects list

            :param objects List of objects to move. (Received by the `self.get_all_list()` method)
            :return None
            """
            for k, object_data in enumerate(objects):
                original_object = object_data.get('Key')
                file_name = original_object.split('/')[-1]
                self.upload_file(f"{out_path}/{file_name}", self.get_bytes_object(original_object))
                self.remove_objects([original_object])
                logger.info(f"Moved {k + 1}/{len(objects)}")

        objects_list = [] if objects_list is None else objects_list
        if in_path:
            objects_list += self.get_list_all_objects(path=in_path)
        objects_list = list(set(objects_list))
        move_objects_list(objects_list)

    def remove_objects(self, remove_objects: list) -> Optional[bool]:
        """
        Remove objects from bucket

        :param remove_objects Data to remove from bucket.
            Example: ['object_name_1', 'object_name_2', ...]

        :return Execution result. True/False
        """
        objects = [{'Key': obj} for obj in remove_objects]
        logger.debug(f'Try deleting objects: {remove_objects}')
        try:
            self.s3.delete_objects(Bucket=self.bucket_name, Delete={'Objects': objects})
        except ClientError as exc:
            return self.__validate_client_error(exc)
        logger.debug('Deleted successfully.')
        return True

    @staticmethod
    def __validate_client_error(client_error: ClientError):
        """ Validate error """
        operation_map = {
            'DeleteObjects': 'Error deleting the list of objects',
            'PutObject': 'Error upload file to bucket',
            'GetObject': 'Error! The specified key does not exist'
        }

        message_code = client_error.response['Error']['Code']
        status_code = client_error.response['ResponseMetadata']['HTTPStatusCode']
        error_msg = operation_map.get(client_error.operation_name, 'Unknown error')

        logger.error(f"{error_msg}. {message_code}/HTTP {status_code}.")



from setuptools import setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as req_file:
    requirements = req_file.read().split('\n')

setup(
    name='maximshumilo_tools',
    version='0.2.3',
    packages=['ms_tools', 'ms_tools.flask', 'ms_tools.object_storage'],
    url='https://t.me/MaximShumilo',
    license='',
    author='Maxim Shumilo',
    author_email='shumilo.mk@gmail.com',
    install_requires=requirements,
    include_package_data=True,
    long_description=long_description,
)
